@@ -114,9 +114,15 @@ import Data.Monoid (Alt(..))
 import Control.Applicative
 import Control.Monad
 import Control.Monad.Fix
+import qualified Control.Monad.ST as Strict
 import qualified Data.Monoid as Monoid
 import Data.List.NonEmpty
-#if MIN_VERSION_base(4,4,0) && !defined(mingw32_HOST_OS) && !defined(ghcjs_HOST_OS)
+#if MIN_VERSION_base(4,6,0)
+import Data.Ord (Down(..))
+#else
+import GHC.Exts (Down(..))
+#endif
+#if MIN_VERSION_base(4,4,0) && !defined(mingw32_HOST_OS) && !defined(ghcjs_HOST_OS) && !defined(ETA_VERSION)
 import GHC.Event
 #endif
 
@@ -238,13 +244,19 @@ class Semigroup a where
       f x y
         | even y = f (x <> x) (y `quot` 2)
         | y == 1 = x
-        | otherwise = g (x <> x) (pred y  `quot` 2) x
+        | otherwise = g (x <> x) (y `quot` 2) x        -- See Note [Half of y - 1]
       g x y z
         | even y = g (x <> x) (y `quot` 2) z
         | y == 1 = x <> z
-        | otherwise = g (x <> x) (pred y `quot` 2) (x <> z)
+        | otherwise = g (x <> x) (y `quot` 2) (x <> z) -- See Note [Half of y - 1]
   {-# INLINE stimes #-}
 
+{- Note [Half of y - 1]
+   ~~~~~~~~~~~~~~~~~~~~~
+   Since y is guaranteed to be odd and positive here,
+   half of y - 1 can be computed as y `quot` 2, optimising subtraction away.
+-}
+
 -- | A generalization of 'Data.List.cycle' to an arbitrary 'Semigroup'.
 -- May fail to terminate for some values in some semigroups.
 cycle1 :: Semigroup m => m -> m
@@ -352,6 +364,14 @@ instance Num a => Semigroup (Product a) where
 #endif
   stimes n (Product a) = Product (a ^ n)
 
+instance Semigroup a => Semigroup (Down a) where
+#ifdef USE_COERCE
+  (<>) = coerce ((<>) :: a -> a -> a)
+#else
+  Down a <> Down b = Down (a <> b)
+#endif
+  stimes n (Down a) = Down (stimes n a)
+
 -- | This is a valid definition of 'stimes' for a 'Monoid'.
 --
 -- Unlike the default definition of 'stimes', it is defined for 0
@@ -365,11 +385,11 @@ stimesMonoid n x0 = case compare n 0 of
       f x y
         | even y = f (x `mappend` x) (y `quot` 2)
         | y == 1 = x
-        | otherwise = g (x `mappend` x) (pred y  `quot` 2) x
+        | otherwise = g (x `mappend` x) (y  `quot` 2) x              -- See Note [Half of y - 1]
       g x y z
         | even y = g (x `mappend` x) (y `quot` 2) z
         | y == 1 = x `mappend` z
-        | otherwise = g (x `mappend` x) (pred y `quot` 2) (x `mappend` z)
+        | otherwise = g (x `mappend` x) (y `quot` 2) (x `mappend` z) -- See Note [Half of y - 1]
 
 -- | This is a valid definition of 'stimes' for an idempotent 'Monoid'.
 --
@@ -1212,7 +1232,14 @@ instance Semigroup a => Semigroup (Tagged s a) where
 instance Semigroup a => Semigroup (IO a) where
     (<>) = liftA2 (<>)
 
-#if !defined(mingw32_HOST_OS) && !defined(ghcjs_HOST_OS)
+instance Semigroup a => Semigroup (Strict.ST s a) where
+#if MIN_VERSION_base(4,4,0)
+    (<>) = liftA2 (<>)
+#else
+    (<>) = liftM2 (<>) -- No Applicative instance for ST on GHC 7.0
+#endif
+
+#if !defined(mingw32_HOST_OS) && !defined(ghcjs_HOST_OS) && !defined(ETA_VERSION)
 # if MIN_VERSION_base(4,4,0)
 instance Semigroup Event where
     (<>) = mappend
